@@ -12,6 +12,7 @@
 #include "QXmppE2eeMetadata.h"
 #include "QXmppStanza_p.h"
 #include "QXmppUtils.h"
+#include "QXmppUtils_p.h"
 
 #include <QDateTime>
 #include <QDomElement>
@@ -83,54 +84,54 @@ QString conditionToString(QXmppStanza::Error::Condition condition)
 
 std::optional<QXmppStanza::Error::Condition> conditionFromString(const QString &string)
 {
-    if (string == "bad-request") {
+    if (string == u"bad-request") {
         return QXmppStanza::Error::BadRequest;
-    } else if (string == "conflict") {
+    } else if (string == u"conflict") {
         return QXmppStanza::Error::Conflict;
-    } else if (string == "feature-not-implemented") {
+    } else if (string == u"feature-not-implemented") {
         return QXmppStanza::Error::FeatureNotImplemented;
-    } else if (string == "forbidden") {
+    } else if (string == u"forbidden") {
         return QXmppStanza::Error::Forbidden;
-    } else if (string == "gone") {
+    } else if (string == u"gone") {
         return QXmppStanza::Error::Gone;
-    } else if (string == "internal-server-error") {
+    } else if (string == u"internal-server-error") {
         return QXmppStanza::Error::InternalServerError;
-    } else if (string == "item-not-found") {
+    } else if (string == u"item-not-found") {
         return QXmppStanza::Error::ItemNotFound;
-    } else if (string == "jid-malformed") {
+    } else if (string == u"jid-malformed") {
         return QXmppStanza::Error::JidMalformed;
-    } else if (string == "not-acceptable") {
+    } else if (string == u"not-acceptable") {
         return QXmppStanza::Error::NotAcceptable;
-    } else if (string == "not-allowed") {
+    } else if (string == u"not-allowed") {
         return QXmppStanza::Error::NotAllowed;
-    } else if (string == "not-authorized") {
+    } else if (string == u"not-authorized") {
         return QXmppStanza::Error::NotAuthorized;
-    } else if (string == "payment-required") {
+    } else if (string == u"payment-required") {
         QT_WARNING_PUSH
         QT_WARNING_DISABLE_DEPRECATED
         return QXmppStanza::Error::PaymentRequired;
         QT_WARNING_POP
-    } else if (string == "policy-violation") {
+    } else if (string == u"policy-violation") {
         return QXmppStanza::Error::PolicyViolation;
-    } else if (string == "recipient-unavailable") {
+    } else if (string == u"recipient-unavailable") {
         return QXmppStanza::Error::RecipientUnavailable;
-    } else if (string == "redirect") {
+    } else if (string == u"redirect") {
         return QXmppStanza::Error::Redirect;
-    } else if (string == "registration-required") {
+    } else if (string == u"registration-required") {
         return QXmppStanza::Error::RegistrationRequired;
-    } else if (string == "remote-server-not-found") {
+    } else if (string == u"remote-server-not-found") {
         return QXmppStanza::Error::RemoteServerNotFound;
-    } else if (string == "remote-server-timeout") {
+    } else if (string == u"remote-server-timeout") {
         return QXmppStanza::Error::RemoteServerTimeout;
-    } else if (string == "resource-constraint") {
+    } else if (string == u"resource-constraint") {
         return QXmppStanza::Error::ResourceConstraint;
-    } else if (string == "service-unavailable") {
+    } else if (string == u"service-unavailable") {
         return QXmppStanza::Error::ServiceUnavailable;
-    } else if (string == "subscription-required") {
+    } else if (string == u"subscription-required") {
         return QXmppStanza::Error::SubscriptionRequired;
-    } else if (string == "undefined-condition") {
+    } else if (string == u"undefined-condition") {
         return QXmppStanza::Error::UndefinedCondition;
-    } else if (string == "unexpected-request") {
+    } else if (string == u"unexpected-request") {
         return QXmppStanza::Error::UnexpectedRequest;
     }
     return std::nullopt;
@@ -285,15 +286,15 @@ void QXmppExtendedAddress::parse(const QDomElement &element)
 
 void QXmppExtendedAddress::toXml(QXmlStreamWriter *xmlWriter) const
 {
-    xmlWriter->writeStartElement(QStringLiteral("address"));
+    xmlWriter->writeStartElement(QSL65("address"));
     if (d->delivered) {
-        xmlWriter->writeAttribute(QStringLiteral("delivered"), QStringLiteral("true"));
+        xmlWriter->writeAttribute(QSL65("delivered"), QStringLiteral("true"));
     }
     if (!d->description.isEmpty()) {
-        xmlWriter->writeAttribute(QStringLiteral("desc"), d->description);
+        xmlWriter->writeAttribute(QSL65("desc"), d->description);
     }
-    xmlWriter->writeAttribute(QStringLiteral("jid"), d->jid);
-    xmlWriter->writeAttribute(QStringLiteral("type"), d->type);
+    xmlWriter->writeAttribute(QSL65("jid"), d->jid);
+    xmlWriter->writeAttribute(QSL65("type"), d->type);
     xmlWriter->writeEndElement();
 }
 /// \endcond
@@ -557,9 +558,7 @@ void QXmppStanza::Error::parse(const QDomElement &errorElement)
     d->type = typeFromString(errorElement.attribute(QStringLiteral("type"))).value_or(NoType);
     d->by = errorElement.attribute(QStringLiteral("by"));
 
-    for (auto element = errorElement.firstChildElement();
-         !element.isNull();
-         element = element.nextSiblingElement()) {
+    for (const auto &element : iterChildElements(errorElement)) {
         if (element.namespaceURI() == ns_stanza) {
             if (element.tagName() == QStringLiteral("text")) {
                 d->text = element.text();
@@ -599,19 +598,19 @@ void QXmppStanza::Error::toXml(QXmlStreamWriter *writer) const
         return;
     }
 
-    writer->writeStartElement(QStringLiteral("error"));
-    helperToXmlAddAttribute(writer, QStringLiteral("by"), d->by);
+    writer->writeStartElement(QSL65("error"));
+    writeOptionalXmlAttribute(writer, u"by", d->by);
     if (d->type != NoType) {
-        writer->writeAttribute(QStringLiteral("type"), typeToString(d->type));
+        writer->writeAttribute(QSL65("type"), typeToString(d->type));
     }
 
     if (d->code > 0) {
-        helperToXmlAddAttribute(writer, QStringLiteral("code"), QString::number(d->code));
+        writeOptionalXmlAttribute(writer, u"code", QString::number(d->code));
     }
 
     if (d->condition != NoCondition) {
         writer->writeStartElement(conditionToString(d->condition));
-        writer->writeDefaultNamespace(ns_stanza);
+        writer->writeDefaultNamespace(toString65(ns_stanza));
 
         // redirection URI
         if (!d->redirectionUri.isEmpty() && (d->condition == Gone || d->condition == Redirect)) {
@@ -621,24 +620,24 @@ void QXmppStanza::Error::toXml(QXmlStreamWriter *writer) const
         writer->writeEndElement();
     }
     if (!d->text.isEmpty()) {
-        writer->writeStartElement(QStringLiteral("text"));
-        writer->writeAttribute(QStringLiteral("xml:lang"), QStringLiteral("en"));
-        writer->writeDefaultNamespace(ns_stanza);
+        writer->writeStartElement(QSL65("text"));
+        writer->writeAttribute(QSL65("xml:lang"), QStringLiteral("en"));
+        writer->writeDefaultNamespace(toString65(ns_stanza));
         writer->writeCharacters(d->text);
         writer->writeEndElement();
     }
 
     // XEP-0363: HTTP File Upload
     if (d->fileTooLarge) {
-        writer->writeStartElement(QStringLiteral("file-too-large"));
-        writer->writeDefaultNamespace(ns_http_upload);
-        helperToXmlAddTextElement(writer, QStringLiteral("max-file-size"),
-                                  QString::number(d->maxFileSize));
+        writer->writeStartElement(QSL65("file-too-large"));
+        writer->writeDefaultNamespace(toString65(ns_http_upload));
+        writeXmlTextElement(writer, u"max-file-size",
+                            QString::number(d->maxFileSize));
         writer->writeEndElement();
     } else if (!d->retryDate.isNull() && d->retryDate.isValid()) {
-        writer->writeStartElement(QStringLiteral("retry"));
-        writer->writeDefaultNamespace(ns_http_upload);
-        writer->writeAttribute(QStringLiteral("stamp"),
+        writer->writeStartElement(QSL65("retry"));
+        writer->writeDefaultNamespace(toString65(ns_http_upload));
+        writer->writeAttribute(QSL65("stamp"),
                                QXmppUtils::datetimeToString(d->retryDate));
         writer->writeEndElement();
     }
@@ -1015,17 +1014,17 @@ void QXmppStanza::generateAndSetNextId()
 {
     // get back
     ++s_uniqeIdNo;
-    d->id = "qxmpp" + QString::number(s_uniqeIdNo);
+    d->id = u"qxmpp" + QString::number(s_uniqeIdNo);
 }
 
 void QXmppStanza::parse(const QDomElement &element)
 {
-    d->from = element.attribute("from");
-    d->to = element.attribute("to");
-    d->id = element.attribute("id");
-    d->lang = element.attribute("lang");
+    d->from = element.attribute(QStringLiteral("from"));
+    d->to = element.attribute(QStringLiteral("to"));
+    d->id = element.attribute(QStringLiteral("id"));
+    d->lang = element.attribute(QStringLiteral("lang"));
 
-    QDomElement errorElement = element.firstChildElement("error");
+    QDomElement errorElement = firstChildElement(element, u"error");
     if (!errorElement.isNull()) {
         Error error;
         error.parse(errorElement);
@@ -1033,9 +1032,7 @@ void QXmppStanza::parse(const QDomElement &element)
     }
 
     // XEP-0033: Extended Stanza Addressing
-    for (auto addressElement = element.firstChildElement("addresses").firstChildElement("address");
-         !addressElement.isNull();
-         addressElement = addressElement.nextSiblingElement("address")) {
+    for (const auto &addressElement : iterChildElements(firstChildElement(element, u"addresses"), u"address")) {
         QXmppExtendedAddress address;
         address.parse(addressElement);
         if (address.isValid()) {
@@ -1048,8 +1045,8 @@ void QXmppStanza::extensionsToXml(QXmlStreamWriter *xmlWriter, QXmpp::SceMode sc
 {
     // XEP-0033: Extended Stanza Addressing
     if (sceMode & QXmpp::ScePublic && !d->extendedAddresses.isEmpty()) {
-        xmlWriter->writeStartElement("addresses");
-        xmlWriter->writeDefaultNamespace(ns_extended_addressing);
+        xmlWriter->writeStartElement(QSL65("addresses"));
+        xmlWriter->writeDefaultNamespace(toString65(ns_extended_addressing));
         for (const auto &address : d->extendedAddresses) {
             address.toXml(xmlWriter);
         }
