@@ -1,4 +1,5 @@
 // SPDX-FileCopyrightText: 2009 Manjeet Dahiya <manjeetdahiya@gmail.com>
+// SPDX-FileCopyrightText: 2019 Linus Jahn <lnj@kaidan.im>
 //
 // SPDX-License-Identifier: LGPL-2.1-or-later
 
@@ -16,20 +17,25 @@
 #include "QXmppLogger.h"
 #include "QXmppMessage.h"
 #include "QXmppMessageHandler.h"
-#include "QXmppOutgoingClient.h"
 #include "QXmppPacket_p.h"
 #include "QXmppPromise.h"
 #include "QXmppRosterManager.h"
+#include "QXmppStreamManagement_p.h"
 #include "QXmppTask.h"
 #include "QXmppTlsManager_p.h"
 #include "QXmppUtils.h"
 #include "QXmppVCardManager.h"
 #include "QXmppVersionManager.h"
 
+#include "XmppSocket.h"
+
+#include <chrono>
+
 #include <QDomElement>
 #include <QSslSocket>
 #include <QTimer>
 
+using namespace std::chrono_literals;
 using namespace QXmpp::Private;
 using MessageEncryptResult = QXmppE2eeExtension::MessageEncryptResult;
 using IqEncryptResult = QXmppE2eeExtension::IqEncryptResult;
@@ -37,8 +43,8 @@ using IqDecryptResult = QXmppE2eeExtension::IqDecryptResult;
 
 static bool isIqResponse(const QDomElement &el)
 {
-    auto type = el.attribute("type");
-    return el.tagName() == "iq" && (type == "result" || type == "error");
+    auto type = el.attribute(QStringLiteral("type"));
+    return el.tagName() == u"iq" && (type == u"result" || type == u"error");
 }
 
 /// \cond
@@ -59,22 +65,22 @@ void QXmppClientPrivate::addProperCapability(QXmppPresence &presence)
 {
     auto *ext = q->findExtension<QXmppDiscoveryManager>();
     if (ext) {
-        presence.setCapabilityHash("sha-1");
+        presence.setCapabilityHash(QStringLiteral("sha-1"));
         presence.setCapabilityNode(ext->clientCapabilitiesNode());
         presence.setCapabilityVer(ext->capabilities().verificationString());
     }
 }
 
-int QXmppClientPrivate::getNextReconnectTime() const
+std::chrono::milliseconds QXmppClientPrivate::getNextReconnectTime() const
 {
     if (reconnectionTries < 5) {
-        return 10 * 1000;
+        return 10s;
     } else if (reconnectionTries < 10) {
-        return 20 * 1000;
+        return 20s;
     } else if (reconnectionTries < 15) {
-        return 40 * 1000;
+        return 40s;
     } else {
-        return 60 * 1000;
+        return 60s;
     }
 }
 
@@ -82,41 +88,66 @@ QStringList QXmppClientPrivate::discoveryFeatures()
 {
     return {
         // XEP-0004: Data Forms
-        ns_data,
+        ns_data.toString(),
         // XEP-0059: Result Set Management
-        ns_rsm,
+        ns_rsm.toString(),
         // XEP-0066: Out of Band Data
-        ns_oob,
+        ns_oob.toString(),
         // XEP-0071: XHTML-IM
-        ns_xhtml_im,
+        ns_xhtml_im.toString(),
         // XEP-0085: Chat State Notifications
-        ns_chat_states,
+        ns_chat_states.toString(),
         // XEP-0115: Entity Capabilities
-        ns_capabilities,
+        ns_capabilities.toString(),
         // XEP-0199: XMPP Ping
-        ns_ping,
+        ns_ping.toString(),
         // XEP-0249: Direct MUC Invitations
-        ns_conference,
+        ns_conference.toString(),
         // XEP-0308: Last Message Correction
-        ns_message_correct,
+        ns_message_correct.toString(),
         // XEP-0333: Chat Markers
-        ns_chat_markers,
+        ns_chat_markers.toString(),
         // XEP-0334: Message Processing Hints
-        ns_message_processing_hints,
+        ns_message_processing_hints.toString(),
         // XEP-0359: Unique and Stable Stanza IDs
-        ns_sid,
+        ns_sid.toString(),
         // XEP-0367: Message Attaching
-        ns_message_attaching,
+        ns_message_attaching.toString(),
         // XEP-0380: Explicit Message Encryption
-        ns_eme,
+        ns_eme.toString(),
         // XEP-0382: Spoiler messages
-        ns_spoiler,
+        ns_spoiler.toString(),
         // XEP-0428: Fallback Indication
-        ns_fallback_indication,
+        ns_fallback_indication.toString(),
         // XEP-0444: Message Reactions
-        ns_reactions,
+        ns_reactions.toString(),
     };
 }
+
+void QXmppClientPrivate::onErrorOccurred(const QString &text, const QXmppOutgoingClient::ConnectionError &err, QXmppClient::Error oldError)
+{
+    if (q->configuration().autoReconnectionEnabled()) {
+        if (oldError == QXmppClient::XmppStreamError) {
+            // if we receive a resource conflict, inhibit reconnection
+            if (stream->xmppStreamError() == QXmppStanza::Error::Conflict) {
+                receivedConflict = true;
+            }
+        } else if (oldError == QXmppClient::SocketError && !receivedConflict) {
+            // schedule reconnect
+            reconnectionTimer->start(getNextReconnectTime());
+        } else if (oldError == QXmppClient::KeepAliveError) {
+            // if we got a keepalive error, reconnect in one second
+            reconnectionTimer->start(1s);
+        }
+    }
+
+    // notify managers
+    Q_EMIT q->error(oldError);
+    Q_EMIT q->errorOccurred(QXmppError {
+        text,
+        visit([](const auto &value) { return std::any(value); }, err),
+    });
+}
 /// \endcond
 
 namespace QXmpp::Private::StanzaPipeline {
@@ -154,7 +185,7 @@ bool process(QXmppClient *client, const QList<QXmppClientExtension *> &extension
 
 bool process(QXmppClient *client, const QList<QXmppClientExtension *> &extensions, QXmppE2eeExtension *e2eeExt, const QDomElement &element)
 {
-    if (element.tagName() != "message") {
+    if (element.tagName() != u"message") {
         return false;
     }
     QXmppMessage message;
@@ -221,14 +252,15 @@ QXmppClient::QXmppClient(InitialExtensions initialExtensions, QObject *parent)
     connect(d->stream->socket(), &QAbstractSocket::stateChanged,
             this, &QXmppClient::_q_socketStateChanged);
 
-    connect(d->stream, &QXmppStream::connected,
+    connect(d->stream, &QXmppOutgoingClient::connected,
             this, &QXmppClient::_q_streamConnected);
 
-    connect(d->stream, &QXmppStream::disconnected,
+    connect(d->stream, &QXmppOutgoingClient::disconnected,
             this, &QXmppClient::_q_streamDisconnected);
 
-    connect(d->stream, &QXmppOutgoingClient::error,
-            this, &QXmppClient::_q_streamError);
+    connect(d->stream, &QXmppOutgoingClient::errorOccurred, this, [this](const auto &text, const auto &error, auto oldError) {
+        d->onErrorOccurred(text, error, oldError);
+    });
 
     // reconnection
     d->reconnectionTimer = new QTimer(this);
@@ -277,19 +309,12 @@ QXmppClient::~QXmppClient() = default;
 ///
 
 /// Registers a new \a extension with the client.
-///
-/// \param extension
-
 bool QXmppClient::addExtension(QXmppClientExtension *extension)
 {
     return insertExtension(d->extensions.size(), extension);
 }
 
 /// Registers a new \a extension with the client at the given \a index.
-///
-/// \param index
-/// \param extension
-
 bool QXmppClient::insertExtension(int index, QXmppClientExtension *extension)
 {
     if (d->extensions.contains(extension)) {
@@ -298,20 +323,20 @@ bool QXmppClient::insertExtension(int index, QXmppClientExtension *extension)
     }
 
     extension->setParent(this);
-    extension->setClient(this);
     d->extensions.insert(index, extension);
+    extension->setClient(this);
     return true;
 }
 
+///
 /// Unregisters the given extension from the client. If the extension
 /// is found, it will be destroyed.
 ///
-/// \param extension
-
 bool QXmppClient::removeExtension(QXmppClientExtension *extension)
 {
     if (d->extensions.contains(extension)) {
         d->extensions.removeAll(extension);
+        extension->setClient(nullptr);
         delete extension;
         return true;
     } else {
@@ -341,16 +366,12 @@ void QXmppClient::setEncryptionExtension(QXmppE2eeExtension *extension)
 }
 
 /// Returns a list containing all the client's extensions.
-///
-
-QList<QXmppClientExtension *> QXmppClient::extensions()
+QList<QXmppClientExtension *> QXmppClient::extensions() const
 {
     return d->extensions;
 }
 
 /// Returns a modifiable reference to the current configuration of QXmppClient.
-/// \return Reference to the QXmppClient's configuration for the connection.
-
 QXmppConfiguration &QXmppClient::configuration()
 {
     return d->stream->configuration();
@@ -371,7 +392,7 @@ void QXmppClient::connectToServer(const QXmppConfiguration &config,
 {
     // reset package cache from last connection
     if (d->stream->configuration().jidBare() != config.jidBare()) {
-        d->stream->resetPacketCache();
+        d->stream->streamAckManager().resetCache();
     }
 
     d->stream->configuration() = config;
@@ -381,11 +402,12 @@ void QXmppClient::connectToServer(const QXmppConfiguration &config,
     d->stream->connectToHost();
 }
 
+///
 /// Overloaded function to simply connect to an XMPP server with a JID and password.
 ///
 /// \param jid JID for the account.
 /// \param password Password for the account.
-
+///
 void QXmppClient::connectToServer(const QString &jid, const QString &password)
 {
     QXmppConfiguration config;
@@ -415,7 +437,7 @@ void QXmppClient::connectToServer(const QString &jid, const QString &password)
 ///
 bool QXmppClient::sendPacket(const QXmppNonza &packet)
 {
-    return d->stream->sendPacket(packet);
+    return d->stream->streamAckManager().sendPacketCompat(packet);
 }
 
 ///
@@ -445,10 +467,10 @@ QXmppTask<QXmpp::SendResult> QXmppClient::sendSensitive(QXmppStanza &&stanza, co
                                QXmlStreamWriter writer(&xml);
                                message->toXml(&writer, QXmpp::ScePublic);
 
-                               d->stream->send(QXmppPacket(xml, true, std::move(interface)));
+                               d->stream->streamAckManager().send(QXmppPacket(xml, true, std::move(interface)));
                            },
                            [&](std::unique_ptr<QXmppIq> &&iq) {
-                               d->stream->send(QXmppPacket(*iq, std::move(interface)));
+                               d->stream->streamAckManager().send(QXmppPacket(*iq, std::move(interface)));
                            },
                            [&](QXmppError &&error) {
                                interface.finish(std::move(error));
@@ -470,7 +492,7 @@ QXmppTask<QXmpp::SendResult> QXmppClient::sendSensitive(QXmppStanza &&stanza, co
                     std::move(dynamic_cast<QXmppIq &&>(stanza)), params));
         }
     }
-    return d->stream->send(stanza);
+    return d->stream->streamAckManager().send(stanza);
 }
 
 ///
@@ -487,7 +509,7 @@ QXmppTask<QXmpp::SendResult> QXmppClient::sendSensitive(QXmppStanza &&stanza, co
 ///
 QXmppTask<QXmpp::SendResult> QXmppClient::send(QXmppStanza &&stanza, const std::optional<QXmppSendStanzaParams> &)
 {
-    return d->stream->send(stanza);
+    return d->stream->streamAckManager().send(stanza);
 }
 
 ///
@@ -629,6 +651,7 @@ QXmppTask<QXmppClient::EmptyResult> QXmppClient::sendGenericIq(QXmppIq &&iq, con
     });
 }
 
+///
 /// Disconnects the client and the current presence of client changes to
 /// QXmppPresence::Unavailable and status text changes to "Logged out".
 ///
@@ -636,14 +659,13 @@ QXmppTask<QXmppClient::EmptyResult> QXmppClient::sendGenericIq(QXmppIq &&iq, con
 /// QXmppPresence::Available, if you are again calling connectToServer() after
 /// calling the disconnectFromServer() function.
 ///
-
 void QXmppClient::disconnectFromServer()
 {
     // cancel reconnection
     d->reconnectionTimer->stop();
 
     d->clientPresence.setType(QXmppPresence::Unavailable);
-    d->clientPresence.setStatusText("Logged out");
+    d->clientPresence.setStatusText(QStringLiteral("Logged out"));
     if (d->stream->isConnected()) {
         sendPacket(d->clientPresence);
     }
@@ -652,15 +674,12 @@ void QXmppClient::disconnectFromServer()
 }
 
 /// Returns true if the client has authenticated with the XMPP server.
-
 bool QXmppClient::isAuthenticated() const
 {
     return d->stream->isAuthenticated();
 }
 
 /// Returns true if the client is connected to the XMPP server.
-///
-
 bool QXmppClient::isConnected() const
 {
     return d->stream->isConnected();
@@ -690,8 +709,8 @@ void QXmppClient::setActive(bool active)
 {
     if (active != d->isActive && isConnected() && d->stream->isClientStateIndicationEnabled()) {
         d->isActive = active;
-        QString packet = "<%1 xmlns='%2'/>";
-        d->stream->sendData(packet.arg(active ? "active" : "inactive", ns_csi).toUtf8());
+        QStringView packet = u"<%1 xmlns='%2'/>";
+        d->stream->xmppSocket().sendData(packet.arg(active ? u"active" : u"inactive", ns_csi).toUtf8());
     }
 }
 
@@ -705,8 +724,8 @@ void QXmppClient::setActive(bool active)
 ///
 QXmppClient::StreamManagementState QXmppClient::streamManagementState() const
 {
-    if (d->stream->isStreamManagementEnabled()) {
-        if (d->stream->isStreamResumed()) {
+    if (d->stream->c2sStreamManager().enabled()) {
+        if (d->stream->c2sStreamManager().streamResumed()) {
             return ResumedStream;
         }
         return NewStream;
@@ -714,19 +733,7 @@ QXmppClient::StreamManagementState QXmppClient::streamManagementState() const
     return NoStreamManagement;
 }
 
-/// Returns the reference to QXmppRosterManager object of the client.
 ///
-/// \return Reference to the roster object of the connected client. Use this to
-/// get the list of friends in the roster and their presence information.
-///
-/// \deprecated This method is deprecated since QXmpp 1.1. Use
-/// \c QXmppClient::findExtension<QXmppRosterManager>() instead.
-
-QXmppRosterManager &QXmppClient::rosterManager()
-{
-    return *findExtension<QXmppRosterManager>();
-}
-
 /// Utility function to send message to all the resources associated with the
 /// specified bareJid. If there are no resources available, that is the contact
 /// is offline or not present in the roster, it will still send a message to
@@ -740,7 +747,7 @@ QXmppRosterManager &QXmppClient::rosterManager()
 ///
 /// \param bareJid bareJid of the receiving entity
 /// \param message Message string to be sent.
-
+///
 void QXmppClient::sendMessage(const QString &bareJid, const QString &message)
 {
     QXmppRosterManager *rosterManager = findExtension<QXmppRosterManager>();
@@ -772,13 +779,12 @@ QXmppClient::State QXmppClient::state() const
 }
 
 /// Returns the client's current presence.
-///
-
 QXmppPresence QXmppClient::clientPresence() const
 {
     return d->clientPresence;
 }
 
+///
 /// Changes the presence of the connected client.
 ///
 /// The connection to the server will be updated accordingly:
@@ -791,7 +797,6 @@ QXmppPresence QXmppClient::clientPresence() const
 ///
 /// \param presence QXmppPresence object
 ///
-
 void QXmppClient::setClientPresence(const QXmppPresence &presence)
 {
     d->clientPresence = presence;
@@ -816,65 +821,37 @@ void QXmppClient::setClientPresence(const QXmppPresence &presence)
 }
 
 /// Returns the socket error if error() is QXmppClient::SocketError.
-///
-
 QAbstractSocket::SocketError QXmppClient::socketError()
 {
     return d->stream->socket()->error();
 }
 
 /// Returns the human-readable description of the last socket error if error() is QXmppClient::SocketError.
-
 QString QXmppClient::socketErrorString() const
 {
     return d->stream->socket()->errorString();
 }
 
 /// Returns the XMPP stream error if QXmppClient::Error is QXmppClient::XmppStreamError.
-
 QXmppStanza::Error::Condition QXmppClient::xmppStreamError()
 {
     return d->stream->xmppStreamError();
 }
 
-///
-/// Returns the reference to QXmppVCardManager, implementation of \xep{0054}.
-/// http://xmpp.org/extensions/xep-0054.html
-///
-/// \deprecated This method is deprecated since QXmpp 1.1. Use
-/// \c QXmppClient::findExtension<QXmppVCardManager>() instead.
-///
-QXmppVCardManager &QXmppClient::vCardManager()
-{
-    return *findExtension<QXmppVCardManager>();
-}
-
-///
-/// Returns the reference to QXmppVersionManager, implementation of \xep{0092}.
-/// http://xmpp.org/extensions/xep-0092.html
-///
-/// \deprecated This method is deprecated since QXmpp 1.1. Use
-/// \c QXmppClient::findExtension<QXmppVersionManager>() instead.
-///
-QXmppVersionManager &QXmppClient::versionManager()
-{
-    return *findExtension<QXmppVersionManager>();
-}
-
 void QXmppClient::injectIq(const QDomElement &element, const std::optional<QXmppE2eeMetadata> &e2eeMetadata)
 {
-    if (element.tagName() != "iq") {
+    if (element.tagName() != u"iq") {
         return;
     }
     if (!StanzaPipeline::process(d->extensions, element, e2eeMetadata)) {
-        const auto iqType = element.attribute("type");
-        if (iqType == "get" || iqType == "set") {
+        const auto iqType = element.attribute(QStringLiteral("type"));
+        if (iqType == u"get" || iqType == u"set") {
             // send error IQ
             using Err = QXmppStanza::Error;
 
             QXmppIq iq(QXmppIq::Error);
-            iq.setTo(element.attribute("from"));
-            iq.setId(element.attribute("id"));
+            iq.setTo(element.attribute(QStringLiteral("from")));
+            iq.setId(element.attribute(QStringLiteral("id")));
             const auto errMessage = e2eeMetadata.has_value()
                 ? QStringLiteral("Feature not implemented or not supported with end-to-end encryption.")
                 : QStringLiteral("Feature not implemented.");
@@ -912,7 +889,7 @@ void QXmppClient::_q_elementReceived(const QDomElement &element, bool &handled)
 void QXmppClient::_q_reconnect()
 {
     if (d->stream->configuration().autoReconnectionEnabled()) {
-        debug("Reconnecting to server");
+        debug(QStringLiteral("Reconnecting to server"));
         d->stream->connectToHost();
     }
 }
@@ -924,7 +901,6 @@ void QXmppClient::_q_socketStateChanged(QAbstractSocket::SocketState socketState
 }
 
 /// At connection establishment, send initial presence.
-
 void QXmppClient::_q_streamConnected()
 {
     d->receivedConflict = false;
@@ -936,7 +912,7 @@ void QXmppClient::_q_streamConnected()
     Q_EMIT stateChanged(QXmppClient::ConnectedState);
 
     // send initial presence
-    if (d->stream->isAuthenticated()) {
+    if (d->stream->isAuthenticated() && streamManagementState() != ResumedStream) {
         sendPacket(d->clientPresence);
     }
 }
@@ -948,34 +924,12 @@ void QXmppClient::_q_streamDisconnected()
     Q_EMIT stateChanged(QXmppClient::DisconnectedState);
 }
 
-void QXmppClient::_q_streamError(QXmppClient::Error err)
-{
-    if (d->stream->configuration().autoReconnectionEnabled()) {
-        if (err == QXmppClient::XmppStreamError) {
-            // if we receive a resource conflict, inhibit reconnection
-            if (d->stream->xmppStreamError() == QXmppStanza::Error::Conflict) {
-                d->receivedConflict = true;
-            }
-        } else if (err == QXmppClient::SocketError && !d->receivedConflict) {
-            // schedule reconnect
-            d->reconnectionTimer->start(d->getNextReconnectTime());
-        } else if (err == QXmppClient::KeepAliveError) {
-            // if we got a keepalive error, reconnect in one second
-            d->reconnectionTimer->start(1000);
-        }
-    }
-
-    // notify managers
-    Q_EMIT error(err);
-}
-
 QXmppLogger *QXmppClient::logger() const
 {
     return d->logger;
 }
 
 /// Sets the QXmppLogger associated with the current QXmppClient.
-
 void QXmppClient::setLogger(QXmppLogger *logger)
 {
     if (logger != d->logger) {
