@@ -11,6 +11,7 @@
 #include "QXmppPubSubEvent.h"
 #include "QXmppTrustManager.h"
 #include "QXmppUtils.h"
+#include "QXmppUtils_p.h"
 
 #include <QStringBuilder>
 
@@ -250,9 +251,11 @@ void QXmppOmemoDevice::setTrustLevel(TrustLevel trustLevel)
 /// client->addNewExtension<QXmppAtmManager>(trustStorage);
 /// \endcode
 ///
-/// Afterwards, the OMEMO manager using its storage must be added to the client:
+/// Afterwards, the OMEMO manager using its storage must be added to the client and activated to be
+/// used for encryption:
 /// \code
 /// auto *manager = client->addNewExtension<QXmppOmemoManager>(omemoStorage);
+/// client->setEncryptionExtension(manager);
 /// \endcode
 ///
 /// You can set a security policy used by OMEMO.
@@ -358,14 +361,14 @@ QXmppTask<bool> Manager::load()
         if (optionalOwnDevice) {
             d->ownDevice = *optionalOwnDevice;
         } else {
-            debug("Device could not be loaded because it is not stored");
+            debug(QStringLiteral("Device could not be loaded because it is not stored"));
             interface.finish(false);
             return;
         }
 
         const auto &signedPreKeyPairs = omemoData.signedPreKeyPairs;
         if (signedPreKeyPairs.isEmpty()) {
-            warning("Signed Pre keys could not be loaded because none is stored");
+            warning(QStringLiteral("Signed Pre keys could not be loaded because none is stored"));
             interface.finish(false);
             return;
         } else {
@@ -375,7 +378,7 @@ QXmppTask<bool> Manager::load()
 
         const auto &preKeyPairs = omemoData.preKeyPairs;
         if (preKeyPairs.isEmpty()) {
-            warning("Pre keys could not be loaded because none is stored");
+            warning(QStringLiteral("Pre keys could not be loaded because none is stored"));
             interface.finish(false);
             return;
         } else {
@@ -440,7 +443,7 @@ QXmppTask<bool> Manager::setUp()
 ///
 QXmppTask<QByteArray> Manager::ownKey()
 {
-    return d->trustManager->ownKey(ns_omemo_2);
+    return d->trustManager->ownKey(ns_omemo_2.toString());
 }
 
 ///
@@ -459,7 +462,7 @@ QXmppTask<QByteArray> Manager::ownKey()
 ///
 QXmppTask<QHash<QXmpp::TrustLevel, QMultiHash<QString, QByteArray>>> Manager::keys(QXmpp::TrustLevels trustLevels)
 {
-    return d->trustManager->keys(ns_omemo_2, trustLevels);
+    return d->trustManager->keys(ns_omemo_2.toString(), trustLevels);
 }
 
 ///
@@ -479,7 +482,7 @@ QXmppTask<QHash<QXmpp::TrustLevel, QMultiHash<QString, QByteArray>>> Manager::ke
 ///
 QXmppTask<QHash<QString, QHash<QByteArray, QXmpp::TrustLevel>>> Manager::keys(const QList<QString> &jids, QXmpp::TrustLevels trustLevels)
 {
-    return d->trustManager->keys(ns_omemo_2, jids, trustLevels);
+    return d->trustManager->keys(ns_omemo_2.toString(), jids, trustLevels);
 }
 
 ///
@@ -567,8 +570,7 @@ void Manager::setMaximumDevicesPerStanza(int maximum)
 QXmppTask<QVector<Manager::DevicesResult>> Manager::requestDeviceLists(const QList<QString> &jids)
 {
     if (const auto jidsCount = jids.size()) {
-        struct State
-        {
+        struct State {
             int processed = 0;
             int jidsCount = 0;
             QXmppPromise<QVector<Manager::DevicesResult>> interface;
@@ -615,8 +617,7 @@ QXmppTask<QVector<Manager::DevicesResult>> Manager::requestDeviceLists(const QLi
 QXmppTask<QVector<Manager::DevicesResult>> Manager::subscribeToDeviceLists(const QList<QString> &jids)
 {
     if (const auto jidsCount = jids.size()) {
-        struct State
-        {
+        struct State {
             int processed = 0;
             int jidsCount = 0;
             QXmppPromise<QVector<Manager::DevicesResult>> interface;
@@ -761,14 +762,14 @@ QXmppTask<QXmppPubSubManager::Result> Manager::removeContactDevices(const QStrin
     auto future = d->unsubscribeFromDeviceList(jid);
     future.then(this, [=](QXmppPubSubManager::Result result) mutable {
         if (std::holds_alternative<QXmppError>(result)) {
-            warning("Contact '" % jid % "' could not be removed because the device list subscription could not be removed");
+            warning(u"Contact '" + jid + u"' could not be removed because the device list subscription could not be removed");
             interface.finish(std::move(result));
         } else {
             d->devices.remove(jid);
 
             auto future = d->omemoStorage->removeDevices(jid);
             future.then(this, [=]() mutable {
-                auto future = d->trustManager->removeKeys(ns_omemo_2, jid);
+                auto future = d->trustManager->removeKeys(ns_omemo_2.toString(), jid);
                 future.then(this, [=]() mutable {
                     interface.finish(std::move(result));
                     Q_EMIT devicesRemoved(jid);
@@ -863,9 +864,9 @@ QXmppTask<void> Manager::buildMissingSessions(const QList<QString> &jids)
     for (const auto &jid : jids) {
         // Do not exceed the maximum of manageable devices.
         if (devicesCount > d->maximumDevicesPerStanza - devicesCount) {
-            warning("Sessions could not be built for all JIDs because their devices are "
-                    "altogether more than the maximum of manageable devices " %
-                    QString::number(d->maximumDevicesPerStanza) %
+            warning(u"Sessions could not be built for all JIDs because their devices are "
+                    "altogether more than the maximum of manageable devices " +
+                    QString::number(d->maximumDevicesPerStanza) +
                     u" - Use QXmppOmemoManager::setMaximumDevicesPerStanza() to increase the maximum");
             break;
         } else {
@@ -952,7 +953,7 @@ QXmppTask<bool> Manager::resetAll()
 ///
 QXmppTask<void> Manager::setSecurityPolicy(QXmpp::TrustSecurityPolicy securityPolicy)
 {
-    return d->trustManager->setSecurityPolicy(ns_omemo_2, securityPolicy);
+    return d->trustManager->setSecurityPolicy(ns_omemo_2.toString(), securityPolicy);
 }
 
 ///
@@ -962,7 +963,7 @@ QXmppTask<void> Manager::setSecurityPolicy(QXmpp::TrustSecurityPolicy securityPo
 ///
 QXmppTask<QXmpp::TrustSecurityPolicy> Manager::securityPolicy()
 {
-    return d->trustManager->securityPolicy(ns_omemo_2);
+    return d->trustManager->securityPolicy(ns_omemo_2.toString());
 }
 
 ///
@@ -975,7 +976,7 @@ QXmppTask<QXmpp::TrustSecurityPolicy> Manager::securityPolicy()
 ///
 QXmppTask<void> Manager::setTrustLevel(const QMultiHash<QString, QByteArray> &keyIds, QXmpp::TrustLevel trustLevel)
 {
-    return d->trustManager->setTrustLevel(ns_omemo_2, keyIds, trustLevel);
+    return d->trustManager->setTrustLevel(ns_omemo_2.toString(), keyIds, trustLevel);
 }
 
 ///
@@ -990,7 +991,7 @@ QXmppTask<void> Manager::setTrustLevel(const QMultiHash<QString, QByteArray> &ke
 ///
 QXmppTask<QXmpp::TrustLevel> Manager::trustLevel(const QString &keyOwnerJid, const QByteArray &keyId)
 {
-    return d->trustManager->trustLevel(ns_omemo_2, keyOwnerJid, keyId);
+    return d->trustManager->trustLevel(ns_omemo_2.toString(), keyOwnerJid, keyId);
 }
 
 /// \cond
@@ -1109,14 +1110,7 @@ QXmppTask<QXmppE2eeExtension::IqDecryptResult> Manager::decryptIq(const QDomElem
 
 bool QXmppOmemoManager::isEncrypted(const QDomElement &el)
 {
-    for (auto subEl = el.firstChildElement();
-         !subEl.isNull();
-         subEl = subEl.nextSiblingElement()) {
-        if (subEl.tagName() == "encrypted" && subEl.namespaceURI() == ns_omemo_2) {
-            return true;
-        }
-    }
-    return false;
+    return !firstChildElement(el, u"encrypted", ns_omemo_2).isNull();
 }
 
 bool QXmppOmemoManager::isEncrypted(const QXmppMessage &message)
@@ -1127,24 +1121,24 @@ bool QXmppOmemoManager::isEncrypted(const QXmppMessage &message)
 QStringList Manager::discoveryFeatures() const
 {
     return {
-        QString(ns_omemo_2_devices) % "+notify"
+        ns_omemo_2_devices + u"+notify"
     };
 }
 
 bool Manager::handleStanza(const QDomElement &stanza)
 {
-    if (stanza.tagName() != "iq" || !QXmppOmemoIq::isOmemoIq(stanza)) {
+    if (stanza.tagName() != u"iq" || !QXmppOmemoIq::isOmemoIq(stanza)) {
         return false;
     }
 
     // TODO: Queue incoming IQs until OMEMO is initialized
     if (!d->isStarted) {
-        warning("Couldn't decrypt incoming IQ because the manager isn't initialized yet.");
+        warning(QStringLiteral("Couldn't decrypt incoming IQ because the manager isn't initialized yet."));
         return false;
     }
 
-    auto type = stanza.attribute("type");
-    if (type != "get" && type != "set") {
+    auto type = stanza.attribute(QStringLiteral("type"));
+    if (type != u"get" && type != u"set") {
         // ignore incoming result and error IQs (they are handled via Client::sendIq())
         return false;
     }
@@ -1153,7 +1147,7 @@ bool Manager::handleStanza(const QDomElement &stanza)
         if (result) {
             injectIq(result->iq, result->e2eeMetadata);
         } else {
-            warning("Could not decrypt incoming OMEMO IQ.");
+            warning(QStringLiteral("Could not decrypt incoming OMEMO IQ."));
         }
     });
     return true;
@@ -1226,11 +1220,8 @@ bool Manager::handleMessage(const QXmppMessage &message)
 ///
 
 /// \cond
-void Manager::setClient(QXmppClient *client)
+void Manager::onRegistered(QXmppClient *client)
 {
-    QXmppClientExtension::setClient(client);
-    client->setEncryptionExtension(this);
-
     d->trustManager = client->findExtension<QXmppTrustManager>();
     if (!d->trustManager) {
         qFatal("QXmppTrustManager is not available, it must be added to the client before adding QXmppOmemoManager");
@@ -1242,7 +1233,7 @@ void Manager::setClient(QXmppClient *client)
     }
 
     connect(d->trustManager, &QXmppTrustManager::trustLevelsChanged, this, [=](const QHash<QString, QMultiHash<QString, QByteArray>> &modifiedKeys) {
-        const auto &modifiedOmemoKeys = modifiedKeys.value(ns_omemo_2);
+        const auto &modifiedOmemoKeys = modifiedKeys.value(ns_omemo_2.toString());
 
         if (!modifiedOmemoKeys.isEmpty()) {
             Q_EMIT trustLevelsChanged(modifiedOmemoKeys);
@@ -1270,6 +1261,12 @@ void Manager::setClient(QXmppClient *client)
     });
 }
 
+void Manager::onUnregistered(QXmppClient *client)
+{
+    // TODO: Proper clean up of connections (currently no issue because extensions are deleted
+    // on removal)
+}
+
 bool Manager::handlePubSubEvent(const QDomElement &element, const QString &pubSubService, const QString &nodeName)
 {
     if (nodeName == ns_omemo_2_devices && QXmppPubSubEvent<QXmppOmemoDeviceListItem>::isPubSubEvent(element)) {
